/****************************************************************
 *  zx80load.c
 *		ZX80-Programme wie vom Band abspielen
 *		=====================================
 *
 *  02.05.99	Bodo Wenzel	V1.0, quick-and-dirty
 *  2002-04-28	Bodo Wenzel	V1.1, auch Linux-Version, keine
 *				Atari-Version mehr
 *  2002-05-06	Bodo Wenzel	V1.2, echte Maschinen brauchen
 *				doch gepulste Signale
 ****************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if defined(__TOS__)
#error "Sorry, Atari-Version nicht weitergepflegt."
#include <tos.h>
#elif defined(__MSDOS__)
#include <bios.h>
#elif defined(__unix__)
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#else
#error "Anpassungen noetig!"
#endif

void quit(char *msg, int ret) {
  /*-------------------------*/
  puts(msg);
  exit(ret);
}

/* Das Verfahren ist bei allen Betriebssystemen gleich: Die
 * Schnittstelle wird auf 4800-8N1 eingestellt. Da das ZX80-BASIC die
 * Bits "nur" ueber die Summe der zeitlichen Laenge des Pulses
 * identifiziert, reicht es, entsprechende Bytes zu senden.
 */

#if defined(__TOS__)

int serinit(void) {
  /*-------------*/
  Rsconf(2, 0, 0x88, 0x01, 0x05, 0);
  return 1;
}

void txdbyte(int d) {
  /*---------------*/
  int i;

  for (i = 0; i < 8; i++, d <<= 1) {
    if ((d & 0x80) != 0)
      Bconout(1, 0x55);
    Bconout(1, 0xF5);
  }
}

#endif

#if defined(__MSDOS__)

int com;

int serinit(char *par) {
  /*------------------*/
  com = -1;
  if (*par == '\0')
    com = 1;
  else if (sscanf(par, "%d", &com) != 1)
    return 0;

  if (com < 1 || com > 4)
    return 0;

  com--;
  bioscom(0, 0xC3, com);
  if ((bioscom(3, 0, com) & 0x6000) != 0)
    return 1;
  return 0;
}

void txdbyte(int d) {
  /*---------------*/
  int i;

  for (i = 0; i < 8; i++, d <<= 1) {
    if ((d & 0x80) != 0)
      bioscom(1, 0x55, com);
    bioscom(1, 0xF5, com);
  }
}

#endif

#if defined(__unix__)

int            fd;
struct termios old_attr;

void serexit(void) {
  /*--------------*/
  tcsetattr(fd, TCSADRAIN, &old_attr);
  close(fd);
}

int serinit(char *port) {
  /*-------------------*/
  struct termios attr;

  fd = open(port, O_RDWR | O_NOCTTY);
  if (fd < 0)
    return 0;

  tcgetattr(fd, &attr);
  old_attr = attr;
  atexit(serexit);

  cfmakeraw(&attr);
  cfsetospeed(&attr, B4800);
  cfsetispeed(&attr, B4800);
  if (tcsetattr(fd, TCSANOW, &attr) != 0)
    return 0;

  tcflush(fd, TCIOFLUSH);
  return 1;
}

void txdbyte(int d) {
  /*---------------*/
  int i, c;

  tcflush(fd, TCIFLUSH);

  for (i = 0; i < 8; i++, d <<= 1) {
    if ((d & 0x80) != 0) {
      c = 0x55;
      (void)write(fd, &c, 1);
    }
    c = 0xF5;
    (void)write(fd, &c, 1);
  }
}

#endif

typedef unsigned char uchar;

#define START  (0x4000)
#define E_LINE (0x400A)
#define BUFLEN (0xC000L - START)

uchar *buffer;

int main(int argc, char *argv[]) {
  /*----------------------------*/
  FILE *f;
  long e_line, d, l, i, dd;

#if defined(__TOS__)
  if (argc < 2)
    quit("Aufruf: ZX80LOAD <datei>", 1);
#endif

#if defined(__MSDOS__)
  if (argc < 2)
    quit("Aufruf: ZX80LOAD <datei> [<com>]", 1);
#endif

#if defined(__unix__)
  setbuf(stdout, NULL);

  if (argc < 3)
    quit("Aufruf: zx80load <datei> <port>", 1);
#endif

  if ((buffer = (uchar *)malloc((size_t)BUFLEN)) == NULL)
    quit("Speicherueberlauf", 4);

#if defined(__TOS__)
  if (!serinit())
    quit("Kann Schnittstelle nicht initialisieren!", 3);
#endif

#if defined(__MSDOS__)
  if (!serinit(argc >= 3 ? argv[2] : ""))
    quit("Kann Schnittstelle nicht initialisieren!", 3);
#endif

#if defined(__unix__)
  if (!serinit(argv[2]))
    quit("Kann Schnittstelle nicht initialisieren!", 3);
#endif

  if ((f = fopen(argv[1], "rb")) == NULL)
    quit("Kann Datei nicht oeffnen!", 2);

  fseek(f, E_LINE - START, SEEK_SET);
  if ((d = fgetc(f)) != EOF) {
    e_line = d;
    if ((d = fgetc(f)) != EOF) {
      e_line |= d << 8;
      fseek(f, 0, SEEK_END);
      l = ftell(f);
      d = EOF;
      if (e_line > START && e_line - 1 <= START + l) {
	l = e_line - START;
	fseek(f, l - 1, SEEK_SET);
	d = fgetc(f);
      }
    }
  }
  if (d != 0x80)
    quit("Kein ZX80-Programm!", 2);

  fseek(f, 0, SEEK_SET);
  if (fread(buffer, sizeof(uchar), (size_t)l, f) != l)
    quit("Lesefehler mit Datei!", 2);
  fclose(f);

  printf("%ld Bytes:\n", l);

  dd = l >> 6;
  for (d = 0; d < l; d += dd)
    putchar('-');
  putchar('\r');

  for (d = 0, i = 0; i < l; i++) {
    if (i > d) {
      d += dd;
      putchar('*');
    }
    txdbyte(buffer[(size_t)i]);
  }
  putchar('\n');
  return 0;
}

/* === Ende =================================================== */
